import numpy as np
import matplotlib.pyplot as plt

from MeineUtils.General import flatten_list

class LinearRegression():
    def __init__(self, file_path, epochs=50, minibatch_size=20, learning_rate = 0.001, data_shuffle=False, delimiter=',', skip_header=1, loss_function=(lambda y_hat, y: (y_hat-y)**2), stochastic=False):
        self.file_path = file_path
        self.epochs = epochs
        self.minibatch_size = minibatch_size
        self.learning_rate = learning_rate
        self.data_shuffle = data_shuffle
        self.loss_function = loss_function

        self.stochastic = stochastic
        if self.stochastic and self.data_shuffle:
        # if self.stochastic and not self.data_shuffle: 
            print(f'data_shuffle: {self.data_shuffle} => False')
            self.data_shuffle = False
            # print(f'data_shuffle: {self.data_shuffle} => True')
            # self.data_shuffle = True

        if '.csv' in self.file_path:
            self.from_csv(file_path=self.file_path, delimiter=delimiter, skip_header=skip_header)

        self.thetas = np.random.randn(self.X.shape[1]+1, 1)
        self.thetas_all = [self.thetas]  

        self.losses_all = [] 
    
    def from_csv(self, file_path, delimiter=',', skip_header=1):
        self.data = np.genfromtxt(file_path,  delimiter=delimiter, skip_header=skip_header)
        self.max_feature = self.data.shape[0]
        if self.minibatch_size == 'all' or self.minibatch_size > self.max_feature:
            print(f'minibatch_size: {self.minibatch_size} => {self.max_feature}')
            self.minibatch_size = self.max_feature
        self.X = self.data[:,:-1]
        self.y = self.data[:,-1:]
        self.X_b = np.concatenate((np.ones((self.max_feature, 1)), self.X), axis=1)
        return self

    def predict(self, x):
        if not isinstance(x, np.ndarray):
            if not isinstance(x, list):
                x = [x]
            x = np.array(x)
        if len(x) == self.X.shape[1]:
            x = np.insert(arr=x, obj=0, values=1.0)
        return x.dot(self.thetas)

    def main(self):
        for _ in range(self.epochs):
            if self.data_shuffle:
                shuffled_indices = np.random.permutation(self.max_feature)
                X_b_shuffled = self.X_b[shuffled_indices]
                y_shuffled = self.y[shuffled_indices]
            else:
                X_b_shuffled = self.X_b
                y_shuffled = self.y

            _loss = []
            for i in range(0, self.max_feature, self.minibatch_size):
                if self.stochastic:
                    # random_index = np.random.randint(np.maximum(self.max_feature-self.minibatch_size, 1))
                    # xi = X_b_shuffled[random_index:random_index+self.minibatch_size]
                    # yi = y_shuffled[random_index:random_index+self.minibatch_size]
                    random_index = np.random.randint(self.max_feature)
                    xi = X_b_shuffled[random_index:random_index+1]
                    yi = y_shuffled[random_index:random_index+1]
                else:
                    xi = X_b_shuffled[i:i+self.minibatch_size]
                    yi = y_shuffled[i:i+self.minibatch_size]
                
                # compute output 
                output = self.predict(xi)
                
                # loss
                loss = self.loss_function(output, yi)
                
                # derivative of loss
                loss_grd = 2*(output - yi)/self.minibatch_size
                
                # derivative of parameters
                gradients = xi.T.dot(loss_grd)
                
                # update thetas
                self.thetas = self.thetas - self.learning_rate*gradients
                self.thetas_all.append(self.thetas)
                
                # loss mean
                loss_mean = np.sum(loss)/self.minibatch_size
                _loss.append(loss_mean)
            self.losses_all.append(_loss)
        return self
    
    def plot(self, color="r", data_range=500):
        losses_list = flatten_list(self.losses_all)
        if data_range > len(losses_list):
            print(f'data_range: {data_range} => {len(losses_list)}')
            data_range = len(losses_list)
        plt.plot(list(range(data_range)), flatten_list(self.losses_all)[:data_range], color=color)
        plt.show()
        return self


import os
import cv2
import numpy as np

"""
I yoink it from https://blog.paperspace.com/data-augmentation-for-object-detection-rotation-and-shearing/
"""
class RandomRotate(object):
    def __init__(self, angle=None):
        if angle:
            self.angle = angle
        else:
            self.angle = np.random.randint(1, 359)
        
    def get_enclosing_box(self, corners):
        x_ = corners[:,[0,2,4,6]]
        y_ = corners[:,[1,3,5,7]]
        
        xmin = np.min(x_,1).reshape(-1,1)
        ymin = np.min(y_,1).reshape(-1,1)
        xmax = np.max(x_,1).reshape(-1,1)
        ymax = np.max(y_,1).reshape(-1,1)
        
        final = np.hstack((xmin, ymin, xmax, ymax,corners[:,8:]))
        
        return final

    def bbox_area(self, bbox):
        return (bbox[:,2] - bbox[:,0])*(bbox[:,3] - bbox[:,1])

    def clip_box(self, bbox, clip_box, alpha):
        ar_ = (self.bbox_area(bbox))
        x_min = np.maximum(bbox[:,0], clip_box[0]).reshape(-1,1)
        y_min = np.maximum(bbox[:,1], clip_box[1]).reshape(-1,1)
        x_max = np.minimum(bbox[:,2], clip_box[2]).reshape(-1,1)
        y_max = np.minimum(bbox[:,3], clip_box[3]).reshape(-1,1)
        
        bbox = np.hstack((x_min, y_min, x_max, y_max, bbox[:,4:]))
        
        delta_area = ((ar_ - self.bbox_area(bbox))/ar_)
        
        mask = (delta_area < (1 - alpha)).astype(int)
        
        bbox = bbox[mask == 1,:]


        return bbox

    def rotate_box(self, corners,angle,  cx, cy, h, w):
        corners = corners.reshape(-1,2)
        corners = np.hstack((corners, np.ones((corners.shape[0],1), dtype = type(corners[0][0]))))
        
        M = cv2.getRotationMatrix2D((cx, cy), angle, 1.0)
        
        
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])
        
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))
        # adjust the rotation matrix to take into account translation
        M[0, 2] += (nW / 2) - cx
        M[1, 2] += (nH / 2) - cy
        # Prepare the vector to be transformed
        calculated = np.dot(M,corners.T).T
        
        calculated = calculated.reshape(-1,8)
        
        return calculated

    def get_corners(self, bboxes):
        width = (bboxes[:,2] - bboxes[:,0]).reshape(-1,1)
        height = (bboxes[:,3] - bboxes[:,1]).reshape(-1,1)
        
        x1 = bboxes[:,0].reshape(-1,1)
        y1 = bboxes[:,1].reshape(-1,1)
        
        x2 = x1 + width
        y2 = y1 
        
        x3 = x1
        y3 = y1 + height
        
        x4 = bboxes[:,2].reshape(-1,1)
        y4 = bboxes[:,3].reshape(-1,1)
        
        corners = np.hstack((x1,y1,x2,y2,x3,y3,x4,y4))
        
        return corners

    def rotate_im(self, image, angle):
        # grab the dimensions of the image and then determine the
        # centre
        (h, w) = image.shape[:2]
        (cX, cY) = (w // 2, h // 2)

        # grab the rotation matrix (applying the negative of the
        # angle to rotate clockwise), then grab the sine and cosine
        # (i.e., the rotation components of the matrix)
        M = cv2.getRotationMatrix2D((cX, cY), angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])

        # compute the new bounding dimensions of the image
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))

        # adjust the rotation matrix to take into account translation
        M[0, 2] += (nW / 2) - cX
        M[1, 2] += (nH / 2) - cY

        return cv2.warpAffine(image, M, (nW, nH))
            
    def __call__(self, img_path, bboxes, save=True):
        img = cv2.imread(img_path)[:,:,::-1]
        w,h = img.shape[1], img.shape[0]
        cx, cy = w//2, h//2
        img = self.rotate_im(img, self.angle)
        corners = self.get_corners(bboxes)
        corners = np.hstack((corners, bboxes[:,4:]))
        corners[:,:8] = self.rotate_box(corners[:,:8], self.angle, cx, cy, h, w)
        new_bbox = self.get_enclosing_box(corners)
        new_bbox = np.array (new_bbox, dtype = float)
        scale_factor_x = img.shape[1] / w
        scale_factor_y = img.shape[0] / h
        img = cv2.resize(img, (w,h))
        new_bbox[:,:4] /= [scale_factor_x, scale_factor_y, scale_factor_x, scale_factor_y] 
        bboxes  = new_bbox
        bboxes = self.clip_box(bboxes, [0,0,w, h], 0.25)

        split_tup = os.path.splitext(img_path)
        save_name = f'{split_tup[0]}-Rotation_{self.angle}{split_tup[1]}'
        if save:
            cv2.imwrite(save_name, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        return img, bboxes, save_name

class Augmentation():
    def __init__(self, image_path, label_path):
        self.image_path = image_path
        self.label_path = label_path
        self.image = np.asarray(cv2.imread(self.image_path))

    def copy_label_content(self, new_label_path):
        lines = open(self.label_path, "r").readlines()
        with open(new_label_path, "w") as f:
            for line in lines:
                f.write(line)

    def new_augemented_path(self, path, new_part):
        return f'{".".join(path.split(".")[:-1])}-{new_part}.{path.split(".")[-1]}'
    
    def fix(self, num):
        if int(num) - num == 0: num = int(num)
        return num

    def save(self, image, portion, part):
        cv2.imwrite(self.new_augemented_path(path=self.image_path,
                                                new_part=f"{part}_{portion}"), image)
        self.copy_label_content(new_label_path=self.new_augemented_path(path=self.label_path,
                                                                        new_part=f"{part}_{portion}"))

class RandomValuedImpulseNoise(Augmentation):
    def __init__(self, image_path, label_path, portion=None):
        super().__init__(image_path, label_path)
        if portion == None:
            self.portion = round(np.random.uniform(low=0.001, high=0.1), 3)
        else:
            assert 0 <= portion <= 0.1, "Please use portion within [0, 0.1]"
            self.portion = portion
        self.thres = 1 - self.portion

    def execute(self, save=True):
        mask = np.random.rand(*self.image.shape[:2])
        mask = np.concatenate([mask, mask, mask]).reshape(*mask.shape, 3)
        self.image = np.where(self.portion*2>mask, np.random.randint(low=0, high=255, size=1, dtype=int), self.image) 
        # self.image = np.where(self.portion>mask, 0, self.image)
        # self.image = np.where(self.thres<mask, 255, self.image)
        if save: self.save(image=self.image, 
                           portion=self.portion*2, 
                           part=type(self).__name__)

class RandomSaltAndPepper(Augmentation):
    def __init__(self, image_path, label_path, portion=None):
        super().__init__(image_path, label_path)
        if portion == None:
            self.portion = round(np.random.uniform(low=0.001, high=0.1), 3)
        else:
            assert 0 <= portion <= 0.1, "Please use portion within [0, 0.1]"
            self.portion = portion
        self.thres = 1 - self.portion

    def execute(self, save=True):
        mask = np.random.rand(self.image.shape[0], self.image.shape[1])
        mask = np.array([mask.T, mask.T, mask.T]).T
        self.image = np.where(self.portion>mask, 0, self.image) 
        self.image = np.where(self.thres<mask, 255, self.image)
        if save: self.save(image=self.image, 
                           portion=self.portion*2, 
                           part=type(self).__name__)

class RandomBrightness(Augmentation):
    def __init__(self, image_path, label_path, level=None):
        super().__init__(image_path, label_path)
        if level == None:
            self.level = round(np.random.uniform(low=0.1, high=2), 3)
        else:
            assert 0.1 <= level <= 2, "Please use level within [0.1, 1), (1, 2]"
            self.level = round(float(level), 3)
        
        while self.level == 1:
            self.level = round(np.random.uniform(low=0.1, high=2), 3)
    def execute(self, save=True):
        self.image = np.clip(a=(self.image * self.level).astype(int),
                             a_min=0,
                             a_max=255)
        if save: self.save(image=self.image, 
                           portion=self.level, 
                           part=type(self).__name__)


def isfloat(num):
    if int(num) - num == 0:
        return False
    else:
        return True

def isfloatfromstring(num):
    try:
        float(num)
        return True
    except ValueError:
        return False


from .linear_regression import LinearRegression
from .logistic_regression import LogisticRegression

# from .normalization import (
#     MinMaxScaler,
#     StandardScaler
# )

from MeineUtils.MachineLearning.bounding_box import BoundingBoxConverter

from MeineUtils.MachineLearning.augmentation import Augmentation
from MeineUtils.MachineLearning.augmentation import RandomBrightness
from MeineUtils.MachineLearning.augmentation import RandomSaltAndPepper
from MeineUtils.MachineLearning.augmentation import RandomValuedImpulseNoise
from MeineUtils.MachineLearning.augmentation import RandomRotate

# from .augmentation import (
#     RandomRotate,
#     RandomNoise,
#     RandomBrightness
# )

from .NaturalLanguageProcessing import (
    preprocessing
)


import os
import re
import time
import numpy as np

from .decorator import _list

@_list
def flatten_list(alist):
    flattened_list = []
    if isinstance(alist, type(np.array([0]))):
        alist = alist.tolist()
    for element in alist:
        if isinstance(element, list):
            flattened_list.extend(flatten_list(element))
        else:
            flattened_list.append(element)
    return flattened_list

@_list
def path_join(alist):
    if len(alist) != 0:
        temp = flatten_list(alist)
        if temp[0] == '':
            return os.sep + os.path.join(*temp)
        else:
            return os.path.join(*temp)
    else:
        return ''

def split(text, delimiters, maxsplit=0, flags=0, strip=True, keep_delimiter=False, remove_empty=True, join_char=None):
    if keep_delimiter:
        regex_pattern = '|'.join('(?<={})'.format(re.escape(delim)) for delim in delimiters)
    else:
        regex_pattern = '|'.join(map(re.escape, delimiters))

    new = re.split(pattern=regex_pattern,
                   string=text, 
                   maxsplit=maxsplit,
                   flags=flags)

    if strip: new = [element.strip() for element in new]
    if remove_empty: new = [a for a in new if a!='']

    if join_char != None:
        assert keep_delimiter==False, 'Will return the original text'
        new = join_char.join(new)
    
    return new

class Sleep():
    def __init__(self):
        self.seconds = 0
        self.minutes = 0
        self.hours = 0
        self.days = 0
        self.time = ''

    def increasement(self):
        self.seconds += 1

        if self.seconds >= 60:
            self.seconds -= 60
            self.minutes += 1
        
        if self.minutes >= 60:
            self.minutes -= 60
            self.hours += 1

        if self.hours >= 60:
            self.hours -= 60
            self.days += 1     

        self.time = ''

        if 0 < self.days < 10: self.time += f"0{self.days}d"  
        if self.days > 10: self.time += f"{self.days}d" 

        if 0 < self.hours < 10: self.time += f"0{self.hours}h"      
        if self.hours > 10: self.time += f"{self.hours}h"

        if 0 < self.minutes < 10: self.time += f"0{self.minutes}m"  
        if self.minutes > 10: f"{self.minutes}m" 

        if 0 < self.seconds < 10: self.time += f"0{self.seconds}s"
        if self.seconds > 10: self.time += f"{self.seconds}s"

        return self 

    def sleep(object):
        try:
            while(True):
                time.sleep(1)
                object.increasement()
                print("", end=f"\rSleeping... {object.time}")
        except KeyboardInterrupt:
            print(f"\n\nKeyboard Interruption")
            return object
            pass


import numpy as np
import matplotlib.pyplot as plt

from MeineUtils.General import flatten_list

class LogisticRegression():
    def __init__(self, 
                 file_path, 
                 delimiter=',', 
                 skip_header=1, 
                 minibatch_size=32,
                 epochs=100,
                 learning_rate = 0.01,
                 data_shuffle=True,
                 thetas=None,
                 loss_function=(lambda y_hat, y: (-y * np.log(y_hat) - (1 - y) * np.log(1 - y_hat)).mean())):
        self.file_path = file_path
        self.minibatch_size = minibatch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.data_shuffle = data_shuffle
        self.loss_function = loss_function
        if '.csv' in self.file_path:
            self.from_csv(file_path=self.file_path, delimiter=delimiter, skip_header=skip_header)

        if thetas:
            thetas = flatten_list(thetas)
            while(len(thetas) < self.X.shape[1]+1): thetas.append(np.random.randn())
            while(len(thetas) > self.X.shape[1]+1): thetas.pop()
            self.thetas = np.array(thetas).reshape(self.X.shape[1]+1, 1)
        else:
            self.thetas = np.random.randn(self.X.shape[1]+1, 1)

        self.thetas_all = [self.thetas]
        self.losses_all = []
        self.accuracy_all = []
    
    def from_csv(self, file_path, delimiter=',', skip_header=1):
        self.data = np.genfromtxt(file_path,  delimiter=delimiter, skip_header=skip_header)
        self.max_feature = self.data.shape[0]
        if self.minibatch_size == 'all' or self.minibatch_size > self.max_feature:
            print(f'minibatch_size: {self.minibatch_size} => {self.max_feature}')
            self.minibatch_size = self.max_feature
        self.X = self.data[:,:-1]
        self.y = self.data[:,-1:]
        self.X_b = np.concatenate((np.ones((self.max_feature, 1)), self.X), axis=1)
        return self

    def sigmoid_function(self, z):
        return 1 / (1 + np.exp(-z))

    def predict(self, x):    
        return self.sigmoid_function(x.dot(self.thetas))

    def main(self):
        for _ in range(self.epochs):
            if self.data_shuffle:
                shuffled_indices = np.random.permutation(self.max_feature)
                X_b_shuffled     = self.X_b[shuffled_indices]
                y_shuffled       = self.y[shuffled_indices]
            else:
                X_b_shuffled = self.X_b
                y_shuffled = self.y

            _loss = []
            _acc = []
            for i in range(0, self.max_feature, self.minibatch_size):
                xi = X_b_shuffled[i:i+self.minibatch_size]
                yi = y_shuffled[i:i+self.minibatch_size]
            
                # compute output
                output = self.predict(xi)

                # compute loss
                loss = self.loss_function(output, yi)

                # compute gradient
                gradient = np.dot(xi.T, (output - yi)) / self.minibatch_size

                # update
                self.thetas -= self.learning_rate*gradient 
                self.thetas_all.append(self.thetas)

                # loss
                _loss.append(loss)

                # accuracy
                preds = self.predict(xi).round()
                acc = (preds == yi).mean()
                _acc.append(acc)
            self.losses_all.append(_loss)
            self.accuracy_all.append(_acc)
        return self

    def plot(self, color="r", data_range=400):
        losses_list = flatten_list(self.losses_all)
        if data_range > len(losses_list):
            print(f'data_range: {data_range} => {len(losses_list)}')
            data_range = len(losses_list)
        fig, ax = plt.subplots(figsize=(10, 5))
        ax.plot(flatten_list(self.accuracy_all), color="#1abd15", linewidth=3, label='accuracy')
        ax.plot(losses_list, color="#d35400", linewidth=3, label='losses')
        ax.legend()
        fig.show()
        return self


import random
import pandas as pd

from .general import split
from .directory import mkdir
from .general import path_join
from .general import flatten_list

def touch(path):
    path = split(text=path, delimiters=['\\', '/'], remove_empty=False)
    mkdir(path_join(path[:-1]))
    open(path_join(path), 'w')

class TXT():
    def __init__(self, file_name):
        self.file_name = file_name
        self.data = []
    
    def read(self, remove_new_line_char=True, encoding=None):
        for line in open(file=self.file_name, mode='r', encoding=encoding).readlines():
            if remove_new_line_char:
                line = line.replace('\n', '')
            self.data.extend(flatten_list(line))
        return self
    
    def write(self, new_line_char=True, mode='w', encoding=None, file_name=None):
        if file_name != None:
            fn = file_name
        else:
            fn = self.file_name
        with open(file=fn, mode=mode, encoding=encoding) as fp:
            for item in self.data:
                if new_line_char: fp.write(f"{item}\n")
                else: fp.write(f"{item}")
        return self
    
    def add_data(self, data):
        self.data.extend(flatten_list(data))
        return self
    
    def drop_duplicate(self):
        assert len(self.data)==0, 'No data'
        self.data = list(set(self.data))
        return self

    def shuffle(self):
        self.data = random.shuffle(self.data)
        return self

    def sort(self, reverse=False):
        self.data.sort(reverse=reverse)
        return self

class CSV():
    def __init__(self, file_name):
        self.file_name = file_name
        self.read()
    
    def read(self):
        self.data = pd.read_csv(self.file_name)
        return self

    def write(self, file_name=None):
        if file_name != None:
            fn = file_name
        else:
            fn = self.file_name
        self.data.to_csv(fn, index=False)

    def drop_duplicate(self):
        self.data = self.data[~self.data.index.duplicated()]
        return self
    
    def sort(self, by, ascending=True):
        self.data.sort_values(by=by, axis=0, ascending=ascending, inplace=True, kind='quicksort', na_position='last')
        return self


from Math.statistics import Mean
from MeineUtils.Math.statistics import StandardDeviation

def MinMaxScaler(x):
    return (x - min(x)) / (max(x) - min(x))

def StandardScaler(x):
    return (x - Mean(x)) / StandardDeviation(x)


import os

from .decorator import _list
from .general import path_join

@_list
def mkdir(dir_list):
    for dir in dir_list:
        dirs = []
        while True:
            if not os.path.exists(dir) and dir != '':
                arr = os.path.normpath(dir).split(os.sep)
                dirs.insert(0, arr[-1])
                dir = path_join(arr[:-1])
            else:
                for i in range(len(dirs)):
                    os.mkdir(path_join([dir, dirs[:i+1]]))
                break


import re
import json
import emot
from textsearch import TextSearch
from nltk.tokenize import sent_tokenize
from nltk.tokenize import word_tokenize

from MeineUtils.General import path_join

class TextPreProcessing():
    def __init__(self, text):
        self.text = text
        self.punctuations = ['!', '"', '#', '$', '%', '&', '\'', '(', ')', '*', '+', ',', '-', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~']

    def sentence_tokenize(self):
        return sent_tokenize(self.text)
    
    def word_tokenize(self):
        return word_tokenize(self.text)
    
    def lowercase(self):
        return self.text.lower()

    def expand_contractions(self, lang='english'):
        contractions_dict = json.load(open(advanced_path_join(['corpus', 'contractions', f'contractions-{lang}.json'])))
        ts_basic = TextSearch("insensitive", "norm")
        ts_basic.add(contractions_dict)
        self.text = ts_basic.replace(self.text)
        return self
    
    def remove_html(self):
        self.text = re.compile('<.*?>').sub(r' ', self.text)
        return self
        
    def remove_urls(self):
        self.text = re.compile(r'https?://\S+|www\.\S+').sub(r' ', self.text)
        return self

    def remove_unwanted_chars(self, unwanted_chars=None, rule=None, count=0, flags=0):
        if unwanted_chars:
            for token in unwanted_chars:
                self.text = self.text.replace(token, "")
        if rule:
            self.text = re.sub(pattern=rule,
                               repl=' ',
                               string=self.text,
                               count=count,
                               flags=flags)
        return self

    def remove_punctuations(self, exclude=None):
        return self.remove_unwanted_chars(unwanted_chars=self.punctuations if exclude==None else [p for p in self.punctuations if p not in exclude])

    def remove_special_characters(self):
        return self.remove_unwanted_chars(rule=
                                                '[^'
                                                u'a-z'
                                                u'A-z'
                                                u'0-9'
                                                u'\s]')

    def remove_emoji(self):
        return self.remove_unwanted_chars(rule=
                                                "["
                                                u"\U0001F600-\U0001F64F"  # emoticons
                                                u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                                                u"\U0001F680-\U0001F6FF"  # transport & map symbols
                                                u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                                                u"\U00002500-\U00002BEF"  # chinese char
                                                u"\U00002702-\U000027B0"
                                                u"\U00002702-\U000027B0"
                                                u"\U000024C2-\U0001F251"
                                                u"\U0001f926-\U0001f937"
                                                u"\U00010000-\U0010ffff"
                                                u"\u2640-\u2642"
                                                u"\u2600-\u2B55"
                                                u"\u200d"
                                                u"\u23cf"
                                                u"\u23e9"
                                                u"\u231a"
                                                u"\ufe0f"  # dingbats
                                                u"\u3030"
                                                "]+", 
                                          flags=re.UNICODE)
        
    def remove_stopwords(self, lang='english'):
        assert lang in ['english', 'vietnamese']
        STOPWORDS = set([line.replace('\n', '').strip() for line in open(advanced_path_join(['corpus' ,'stopwords', f'stopwords-{lang}.txt']), 'r').readlines()])
        self.text = " ".join([word for word in str(self.text).split() if word not in STOPWORDS])
        return self

    def convert_emoticons(self):
        dict_emoticons = dict(zip(emot.emot().emoticons(self.text)['value'], emot.emot().emoticons(self.text)['mean']))
        res_emoticons =  dict(sorted(dict_emoticons.items(), key = lambda kv:len(kv[1]), reverse=True))
        for emoticon, mean in res_emoticons.items():
            self.text = self.text.replace(emoticon, mean)
        return self

    def convert_emojis(self):
        for emoji, mean in zip(emot.emot().emoji(self.text)['value'], emot.emot().emoji(self.text)['mean']):
            self.text = self.text.replace(emoji, mean.replace(":", ""))
        return self

    def fix_format(self):
        self.text = " ".join(self.text.split()) 
        fix_spaces = re.compile(r'\s*([?!.,]+(?:\s+[?!.,]+)*)\s*')
        self.text = fix_spaces.sub(lambda x: "{} ".format(x.group(1).replace(" ", "")), self.text).strip()

        return self

    def get_text(self):
        return self.text


import platform
import contextlib

def _list(func):
    def wrap(a):
        if isinstance(a, list):
            retval = func(a)
        else:
            retval = func([a])
        return retval
    return wrap

class TryExcept(contextlib.ContextDecorator):
    """
    Example:
    @TryExcept('⚠️ nani the fck: ')
    def calculate(x):
        return x/0
    """
    def __init__(self, msg=''):
        self.msg = msg

    def __enter__(self):
        pass

    def emojis(self, str=''):
        """
        Return platform-dependent emoji-safe version of string
        """
        return str.encode().decode('ascii', 'ignore') if platform.system() == 'Windows' else str

    def __exit__(self, exc_type, value, traceback):
        if value:
            print(self.emojis(f'{self.msg}{value}'))
        return True


import os
import cv2
import numpy as np
from pathlib import Path
from xml.dom import minidom

class BoundingBoxConverter():
    def __init__(self) -> None:
        """ Pascal VOC Format """
        self.xTopLefts = []
        self.yTopLefts = []
        self.xBottomRights = []
        self.yBottomRights = []

        """ COCO Format """
        self.widthBBoxes = []
        self.heightBBoxes = []

        """ YOLO Format """
        self.xCenterScaled = []
        self.yCenterScaled = []
        self.widthScaled = []
        self.heightScaled = []

        """ Image """
        self.width = None
        self.height = None
        self.imagePath = None

        """ Label """
        self.labels = []
        self.classes = []
        self.labelDirectory = None
        self.labelFilePath = None
        self.labelFileName = None

    def addLabelDirectory(self, labelDirectory):
        self.labelDirectory = labelDirectory
        return self

    def addlabelFileName(self, labelFileName):
        self.labelFileName = labelFileName
        return self

    def YOLO_to_PascalVOC(self):
        self.xTopLefts.extend([0 if (x - w / 2) * self.width < 0 else int((x - w / 2) * self.width) + 1 for x, w in zip(self.xCenterScaled, self.widthScaled)])
        self.xBottomRights.extend([self.width - 1 if (x + w / 2) * self.width > self.width - 1 else int((x + w / 2) * self.width) + 1 for x, w in zip(self.xCenterScaled, self.widthScaled)])
        self.yTopLefts.extend([0 if (y - h / 2) * self.height < 0 else int((y - h / 2) * self.height) + 1 for y, h in zip(self.yCenterScaled, self.heightScaled)])
        self.yBottomRights.extend([self.height - 1 if (y + h / 2) * self.height > self.height - 1 else int((y + h / 2) * self.height) + 1 for y, h in zip(self.yCenterScaled, self.heightScaled)])

    def PascalVOC_to_YOLO(self):
        self.xCenterScaled.extend([((x1 + x2)/2.0 - 1) * (1./self.width) for x1, x2 in zip(self.xTopLefts, self.xBottomRights)])
        self.yCenterScaled.extend([(x2 - x1) * (1./self.width) for x1, x2 in zip(self.xTopLefts, self.xBottomRights)])
        self.widthScaled.extend([((y1 + y2)/2.0 - 1) * (1./self.height) for y1, y2 in zip(self.yTopLefts, self.yBottomRights)])
        self.heightScaled.extend([(y2 - y1) * (1./self.width) for y1, y2 in zip(self.yTopLefts, self.yBottomRights)])

    def COCO_to_YOLO(self):
        self.xCenterScaled.extend([(2 * x1 + w)/(2 * self.width) for x1, w in zip(self.xTopLefts, self.widthBBoxes)])
        self.yCenterScaled.extend([(2 * y1 + h)/(2 * self.height) for y1, h in zip(self.yTopLefts, self.heightBBoxes)])
        self.widthScaled.extend([w / self.width for w in self.widthBBoxes])
        self.heightScaled.extend([h / self.height for h in self.heightBBoxes])


    def addBoundingBox(self, format, bbox=None, labelFilePath=None):
        if labelFilePath:
            self.labelFilePath = labelFilePath
            self.labelFileName = Path(labelFilePath).stem

        if format.lower().replace(" ", "") == "yolo":
            if labelFilePath and labelFilePath.endswith(".txt"):
                lines = np.array([line.strip().split(" ") for line in open(labelFilePath, "r").readlines()], dtype=float)
            
            elif bbox and isinstance(bbox, list):
                if all(isinstance(i, list) for i in bbox):
                    lines = np.array(bbox, dtype=float)
                elif all(isinstance(i, str) for i in bbox):
                    lines = np.array([i.strip().split(" ") for i in bbox], dtype=float)

            if len(lines[0]) == 4:
                self.xCenterScaled.extend(lines[:, 0])
                self.yCenterScaled.extend(lines[:, 1])
                self.widthScaled.extend(lines[:, 2])
                self.heightScaled.extend(lines[:, 3])
            elif len(lines[0]) == 5:
                self.labels.extend([int(label) for label in lines[:, 0]])
                self.xCenterScaled.extend(lines[:, 1])
                self.yCenterScaled.extend(lines[:, 2])
                self.widthScaled.extend(lines[:, 3])
                self.heightScaled.extend(lines[:, 4])
        
        elif format.lower().replace(" ", "") == "pascalvoc":
            if labelFilePath and labelFilePath.endswith(".xml"):
                file = minidom.parse(labelFilePath)
                
                self.labels.extend([l.firstChild.data for l in list(file.getElementsByTagName('name'))])
                self.xTopLefts.extend([int(x1.firstChild.data) for x1 in list(file.getElementsByTagName('xmin'))])
                self.yTopLefts.extend([int(y1.firstChild.data) for y1 in list(file.getElementsByTagName('ymin'))])
                self.xBottomRights.extend([int(x2.firstChild.data) for x2 in list(file.getElementsByTagName('xmax'))])
                self.yBottomRights.extend([int(y2.firstChild.data) for y2 in list(file.getElementsByTagName('ymax'))])

                self.width = int(list(file.getElementsByTagName('width'))[0].firstChild.data)
                self.height = int(list(file.getElementsByTagName('height'))[0].firstChild.data)
            
            elif bbox and isinstance(bbox, list):
                if all(isinstance(i, list) for i in bbox):
                    lines = np.array(bbox, dtype=float)
                elif all(isinstance(i, str) for i in bbox):
                    lines = np.array([i.strip().split(" ") for i in bbox], dtype=float)

            if len(lines[0]) == 4:
                self.xTopLefts.extend(lines[:, 0])
                self.yTopLefts.extend(lines[:, 1])
                self.xBottomRights.extend(lines[:, 2])
                self.yBottomRights.extend(lines[:, 3])
            elif len(lines[0]) == 5:
                self.labels.extend([int(label) for label in lines[:, 0]])
                self.xTopLefts.extend(lines[:, 1])
                self.yTopLefts.extend(lines[:, 2])
                self.xBottomRights.extend(lines[:, 3])
                self.yBottomRights.extend(lines[:, 4])

        return self

    def addImageShape(self, imagePath=None, size=None):
        if imagePath:
            img = cv2.imread(imagePath)[:,:,::-1]
            self.width, self.height = img.shape[1], img.shape[0]
            self.imagePath = imagePath

        if size:
            self.width = size[0]
            self.height = size[1]
        
        return self

    def save(self, format):
        if not self.labelFileName and self.imagePath:
            self.labelFileName = '.'.join(os.path.split(self.imagePath)[1].split('.')[:-1])
        elif not self.labelDirectory and self.labelFilePath:
            self.labelDirectory = os.path.dirname(self.labelFilePath)

        if format.lower() == "yolo":
            file = open(f"{os.path.join(self.labelDirectory, self.labelFileName)}.txt", "w")
            file.writelines([f"{l} {x} {y} {w} {h}\n" for l, x, y, w, h in zip(self.labels, self.xCenterScaled, self.yCenterScaled, self.widthScaled, self.heightScaled)])

        elif format.lower() == "pascalvoc":
            pass

    def addClasses(self, classes):
        if isinstance(classes, list):
            self.classes.extend(classes)

        elif isinstance(classes, str) and classes.endswith(".txt"):
            self.classes.extend([line.replace(" ", "").replace("\n", "").split(",") for line in open(classes, "r").readlines()])
            print(self.classes)
        return self

    def retrieveLabelIndex(self, label):
        return self.classes.index(label)

    def retrieveLabelName(self, label):
        return self.classes[label]

    def changeLabelFormat(self, reverse=False):
        assert self.classes, "Please provide classes using .addClasses before continuing"
        if reverse:
            for index, l in enumerate(self.labels):
                self.labels[index] = self.retrieveLabelName(l) 
        else:
            for index, l in enumerate(self.labels):
                self.labels[index] = self.retrieveLabelIndex(l) 

    def export(self, format, save=False):
        if format.lower() == "pascalvoc":
            if not all([self.xTopLefts, self.yTopLefts, self.xBottomRights, self.yBottomRights]):
                if all([self.xCenterScaled, self.yCenterScaled, self.widthScaled, self.heightScaled]):
                    assert all([self.width, self.height]), "Please provide image shape using .addImageShape before continuing"
                    assert (all([self.xCenterScaled, self.yCenterScaled, self.widthScaled, self.heightScaled]))
                    self.YOLO_to_PascalVOC()

            if save:
                if all(isinstance(x, int) for x in self.labels):
                    self.changeLabelFormat(reverse=True)
                self.save(format=format)

            return self, [(x1, y1, x2, y2) for x1, y1, x2, y2 in zip(self.xTopLefts, self.yTopLefts, self.xBottomRights, self.yBottomRights)]

        elif format.lower() == "yolo":
            if not all([self.xCenterScaled, self.yCenterScaled, self.widthScaled, self.heightScaled]):
                if all([self.xTopLefts, self.yTopLefts, self.xBottomRights, self.yBottomRights]):
                    assert all([self.width, self.height]), "Please provide image shape using .addImageShape before continuing"
                    assert all([self.xTopLefts, self.yTopLefts, self.xBottomRights, self.yBottomRights])
                    self.PascalVOC_to_YOLO()

            if save:
                if all(isinstance(x, str) for x in self.labels):
                    self.changeLabelFormat()
                self.save(format=format)

            return self, [(x, y, w, h) for x, y, w, h in zip(self.xCenterScaled, self.yCenterScaled, self.widthScaled, self.heightScaled)]



import os
import codecs

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

DESCRIPTION = 'My utils'

# Setting up
setup(
    name="MeineUtils",
    version='0.0.42',
    author="neko941 (Khoa Nguyen)",
    url='https://github.com/neko941/MeineUtils',
    # author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['pandas', 'emot', 'textsearch', 'nltk', 'numpy', 'opencv-python'],
    keywords=['python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import numpy as np

def ReLU_base(x, alpha1, alpha2):
    return alpha1 * np.minimum(0, x) + alpha2 * np.maximum(0, x)

def ReLU(x):
    """ 
    Rectified Linear Unit
    f(x) = max(0, x)
    """
    return ReLU_base(x, 0, 1)

def PReLU(x):
    """
    Parametric ReLU
    """
    return ReLU_base(x, 0.25, 1)

def LeakyReLU(x):
    return ReLU_base(x, 0.01, 1)


__all__ = ['_list', 'TryExcept', 'mkdir', 'touch', 'TXT', 'CSV', 'flatten_list', 'path_join', 'split', 'Sleep']

from .decorator import (
    _list,
    TryExcept
)

from .directory import (
    mkdir
)

from .file import (
    touch,
    TXT,
    CSV
)

from .general import (
    flatten_list,
    path_join,
    split,
    Sleep
)


from .activation_funciton import (
    ReLU_base,
    ReLU,
    PReLU,
    LeakyReLU
)

from .general import (
    isfloat,
    isfloatfromstring
)

# from .statistics import (
#     Mean,
#     StandardDeviation
# )